import os
import tempfile
import unittest

from plone.reload import xreload

TESTS = os.path.join(os.path.dirname(__file__), 'data')


class TestReload(unittest.TestCase):

    def _create_temp(self):
        # This creates both a temporary file used as a module and also adds
        # an __init__.py to the data subfolder, to temporarily make it a
        # package
        temp = tempfile.NamedTemporaryFile(
            mode='w', suffix='.py', dir=TESTS, delete=False)
        temp.close()
        fd = open(os.path.join(TESTS, '__init__.py'), mode="w")
        fd.write("#")
        fd.close()
        name = os.path.split(temp.name)[-1]
        modulename = 'plone.reload.tests.data.' + name[:-3]
        module = __import__(modulename,
                            fromlist=['plone', 'reload', 'tests', 'data'])
        return temp.name, module

    def reload(self, text=None):
        if text is not None:
            temp = open(self.name, mode='w')
            temp.write(text)
            temp.close()
        self.reloader.reload()

    def setUp(self):
        self.name, self.module = self._create_temp()
        self.reloader = xreload.Reloader(self.module)

    def tearDown(self):
        os.unlink(self.name)
        sfile = self.name + 'c'
        if os.path.isfile(sfile):
            os.unlink(sfile)
        data_init = os.path.join(TESTS, '__init__.py')
        if os.path.isfile(data_init):
            os.unlink(data_init)
        if os.path.isfile(data_init + 'c'):
            os.unlink(data_init + 'c')


class TestReloadModule(TestReload):

    def test_stdlib(self):
        import sre_constants
        r = xreload.Reloader(sre_constants)
        r.reload()

    def test_immutable_constant_added(self):
        self.reload("FOO = 15")
        self.assertEqual(self.module.FOO, 15)

    def test_immutable_constant_changed(self):
        self.reload("FOO = 15")
        self.reload("FOO = 16")
        self.assertEqual(self.module.FOO, 16)

    def test_immutable_constant_removed(self):
        self.reload("FOO = 15")
        self.reload("")
        self.assertTrue(getattr(self.module, 'FOO', None) is None)

    def test_mutable_constant_added(self):
        self.reload("FOO = [1, 2]")
        self.assertEqual(self.module.FOO, [1, 2])

    def test_instance_constant_added(self):
        self.reload("FOO = object()")
        self.assertTrue(isinstance(self.module.FOO, object))

    def test_function_added(self):
        self.reload("def foo(): return True")
        self.assertTrue(self.module.foo())

    def test_class_added(self):
        self.reload(
            "class Foo(object):\n\tdef __init__(self):\n\t\tself.bar = 1")
        self.assertEqual(self.module.Foo().bar, 1)

    def test_import_added(self):
        self.reload("def foo(): pass")
        self.reload("import os\ndef foo(): return os.pathsep")
        self.assertEqual(self.module.foo(), os.pathsep)


class TestReloadFunction(TestReload):

    def test_function_added(self):
        source = "def foo(): return True"
        self.reload(source)
        source = source + "\ndef bar(): return True\n"
        self.reload(source)
        self.assertTrue(self.module.foo())
        self.assertTrue(self.module.bar())

    def test_function_changed(self):
        self.reload("def foo(): return True")
        self.reload("def foo(): return False")
        self.assertFalse(self.module.foo())

    def test_function_changed_to_constant(self):
        self.reload("def foo(): return True")
        self.reload("foo = 15")
        self.assertEqual(self.module.foo, 15)

    def test_function_removed(self):
        source = "def foo(): return True"
        self.reload(source + "\ndef bar(): return True\n")
        self.reload(source)
        self.assertTrue(self.module.foo())
        self.assertTrue(getattr(self.module, 'bar', None) is None)


class TestReloadClass(TestReload):

    def test_class_added(self):
        source = "class Foo(object):\n\tdef __init__(self):\n\t\tself.bar = 1"
        self.reload(source)
        source = source + "\nclass Bar(object): bar = 2"
        self.reload(source)
        self.assertEqual(self.module.Foo().bar, 1)
        self.assertEqual(self.module.Bar().bar, 2)

    def test_class_variable_added(self):
        source = "class Foo(object):\n\tfoo = 1"
        self.reload(source)
        source = source + "\n\tbar = 2"
        self.reload(source)
        self.assertEqual(self.module.Foo().foo, 1)
        self.assertEqual(self.module.Foo().bar, 2)

    def test_class_variable_changed(self):
        self.reload("class Foo(object):\n\tfoo = 1")
        self.reload("class Foo(object):\n\tfoo = 2")
        self.assertEqual(self.module.Foo().foo, 2)

    def test_class_variable_removed(self):
        self.reload("class Foo(object):\n\tfoo = 1")
        self.reload("class Foo(object):\n\tpass")
        # We don't remove any class variables
        self.assertEqual(self.module.Foo().foo, 1)

    def test_class_init_added(self):
        base = "class Foo(object):\n"
        self.reload(base + "\tpass")
        self.reload(base + "\tdef __init__(self):\n\t\tself.bar = 2")
        self.assertEqual(self.module.Foo().bar, 2)

    def test_class_init_changed(self):
        base = "class Foo(object):\n\tdef __init__(self):\n"
        self.reload(base + "\t\tself.bar = 1")
        self.reload(base + "\t\tself.bar = 2")
        self.assertEqual(self.module.Foo().bar, 2)

    def test_class_init_removed(self):
        base = "class Foo(object):\n"
        self.reload(base + "\tdef __init__(self):\n\t\tself.bar = 2")
        self.reload(base + "\tpass")
        # We don't remove anything from a class
        self.assertEqual(self.module.Foo().bar, 2)

    def test_class_method_added(self):
        source = "class Foo(object):\n\tfoo = 1"
        self.reload(source)
        source = source + "\n\tdef bar(self):\n\t\treturn 2"
        self.reload(source)
        self.assertEqual(self.module.Foo().foo, 1)
        self.assertEqual(self.module.Foo().bar(), 2)

    def test_class_method_changed(self):
        base = "class Foo(object):\n\tdef foo(self, a):\n"
        self.reload(base + "\t\treturn a + 1")
        self.reload(base + "\t\treturn a + 2")
        self.assertEqual(self.module.Foo().foo(2), 4)

    def test_class_method_arguments_changed(self):
        base = "class Foo(object):\n"
        self.reload(base + "\tdef foo(self, a):\n\t\treturn a + 1")
        self.reload(base + "\tdef foo(self, a, b):\n\t\treturn a + b")
        self.assertEqual(self.module.Foo().foo(2, 3), 5)

    def test_class_property_added(self):
        base = "class Foo(object):\n"
        self.reload(base + "\tpass")
        self.reload(base + "\t@property\n\tdef foo(self):\n\t\treturn 1")
        self.assertEqual(self.module.Foo().foo, 1)

    def test_class_property_changed(self):
        base = "class Foo(object):\n\t@property\n\tdef foo(self):\n"
        self.reload(base + "\t\treturn 1")
        self.reload(base + "\t\treturn 2")
        self.assertEqual(self.module.Foo().foo, 2)

    def test_class_property_changed_to_method(self):
        base = "class Foo(object):\n"
        self.reload(base + "\t@property\n\tdef foo(self):\n\t\treturn 1")
        self.reload(base + "\n\tdef foo(self):\n\t\treturn 2")
        self.assertEqual(self.module.Foo().foo(), 2)

    def test_class_method_changed_to_property(self):
        base = "class Foo(object):\n"
        self.reload(base + "\n\tdef foo(self):\n\t\treturn 2")
        self.reload(base + "\t@property\n\tdef foo(self):\n\t\treturn 1")
        self.assertEqual(self.module.Foo().foo, 1)

    def test_class_staticmethod_added(self):
        base = "class Foo(object):\n"
        self.reload(base + "\tpass")
        self.reload(base + "\t@staticmethod\n\tdef foo():\n\t\treturn 1")
        self.assertEqual(self.module.Foo.foo(), 1)

    def test_class_staticmethod_changed(self):
        base = "class Foo(object):\n\t@staticmethod\n\tdef foo():\n"
        self.reload(base + "\t\treturn 1")
        self.reload(base + "\t\treturn 2")
        self.assertEqual(self.module.Foo.foo(), 2)

    def test_class_nested_function_changed(self):
        base = "class Foo(object):\n\tdef foo(self, a):\n"
        self.reload(base + "\t\tdef bar(a): return a + a\n\t\treturn bar(a)")
        self.reload(base + "\t\tdef bar(a): return a + 2\n\t\treturn bar(a)")
        self.assertEqual(self.module.Foo().foo(5), 7)

    def test_class_class_instance_changed(self):
        base = ("class Foo(object):\n\tdef f(self): return %s\n"
                "class Bar(object):\n")
        self.reload(base % 1 + "\tfoo = Foo()")
        self.assertEqual(self.module.Bar().foo.f(), 1)
        self.reload(base % 2 + "\tfoo = Foo()")
        self.assertEqual(self.module.Bar().foo.f(), 2)

    def test_class_descriptor_changed(self):
        base = ("class Foo(object):\n"
                "\tdef i(self): return 'Foo'\n"
                "class Bar(object):\n"
                "\tdef i(self): return 'Bar'\n"
                "\tdef __get__(self, instance, type): return Foo()\n"
                "class Baz(object):\n")
        self.reload(base + "\tbar = Bar()")
        self.reload(base + "\tbar = Bar()")
        baz = self.module.Baz
        self.assertEqual(baz.bar.i(), 'Foo')
        self.assertEqual(baz.__dict__.get('bar').i(), 'Bar')


class TestReloadDecorator(TestReload):

    base = """\
def outer(func):
    def inner(self, *args):
        return func(self, *args)
    return inner

class Foo(object):
"""

    def test_class_decorater_added(self):
        self.reload(self.base + "\tdef foo(self, a): return a + 1")
        self.reload(self.base + "\t@outer\n\tdef foo(self, a): return a + 2")
        self.assertEqual(self.module.Foo().foo(2), 4)

    def test_class_decorated_method_changed(self):
        self.reload(self.base + "\t@outer\n\tdef foo(self, a): return a + 1")
        self.reload(self.base + "\t@outer\n\tdef foo(self, a): return a + 2")
        self.assertEqual(self.module.Foo().foo(3), 5)

    def test_class_decorator_changed(self):
        source = self.base + "\t@outer\n\tdef foo(self, a): return a + 1"
        self.reload(source)
        source = source.replace("func(self, *args)", "func(self, 15)")
        self.reload(source)
        self.assertEqual(self.module.Foo().foo(4), 16)

    def test_class_decorator_removed(self):
        self.reload(self.base + "\t@outer\n\tdef foo(self, a): return a + 1")
        self.reload(self.base + "\tdef foo(self, a): return a + 2")
        self.assertEqual(self.module.Foo().foo(5), 7)


class TestReloadInterface(TestReload):

    base = '''\
from zope.interface import Interface
class IFoo(Interface):
    def bar():
        """A true bar."""
'''

    def test_interface_method_added(self):
        self.reload(self.base)
        self.reload(self.base + '\tdef baz():\n\t\t"""Maybe a baz?"""')
        self.assertTrue('bar' in self.module.IFoo.names())
        # Reloading interfaces doesn't work yet at all
        self.assertFalse('baz' in self.module.IFoo.names())



from Products.CMFCore.FSObject import FSObject
from Products.CMFCore.interfaces import ISiteRoot
from Products.CMFCore.utils import getToolByName


def reload_skins(tool):
    counter = 0
    for folder in tool.objectValues():
        for obj in folder.objectValues():
            if isinstance(obj, FSObject):
                parsed = getattr(obj, '_parsed', 0)
                if parsed:
                    obj._parsed = 0
                    counter += 1
    return counter


def reload_template(root):
    counter = 0
    for obj in root.objectValues():
        if ISiteRoot.providedBy(obj):
            tool = getToolByName(obj, 'portal_skins', None)
            if tool is not None:
                counter = reload_skins(tool)
    return counter



import os
import sys

from os.path import abspath
from os.path import isfile

from plone.reload import config
from plone.reload.xreload import Reloader

_marker = object()
MOD_TIMES = dict()


def in_search_path(path):
    if 'site-packages' in path:
        return False
    elif '.egg' in path:
        return False
    return True


def search_modules():
    modules = []
    for name, module in sys.modules.items():
        if module is not None:
            f = getattr(module, '__file__', None)
            # Standard library modules don't have a __file__
            if f is None:
                continue
            f = abspath(f)
            if config.EXCLUDE_SITE_PACKAGES:
                if in_search_path(f):
                    modules.append((f, module))
            else:
                modules.append((f, module))
    return modules


def get_mod_time(path):
    mtime = 0
    # If we have the compiled source, look for the source code change date
    if path.endswith('pyc') or path.endswith('pyo'):
        source = path[:-1]
        if os.path.isfile(source):
            path = source
    # protect against missing and unaccessible files
    if isfile(path):
        mtime = os.stat(path)[8]
    return mtime


def get_mod_times(event=None):
    global MOD_TIMES
    for path, module in search_modules():
        if path not in MOD_TIMES:
            MOD_TIMES[path] = (get_mod_time(path), module)
    return MOD_TIMES


def check_mod_times():
    changed = []
    for path, (time, module) in get_mod_times().items():
        newtime = get_mod_time(path)
        if time < newtime:
            changed.append((path, newtime, module))
    return changed


def reload_code():
    global MOD_TIMES
    reloaded = []
    for path, time, module in check_mod_times():
        r = Reloader(module)
        module = r.reload()
        MOD_TIMES[path] = (time, module)
        reloaded.append(path)
    return reloaded



import os
import sys
import types
import unittest

TESTS = os.path.dirname(__file__)


class TestSearch(unittest.TestCase):

    def test_in_search_path(self):
        from plone.reload.code import in_search_path
        self.assertTrue(in_search_path('/home/foo/plone.reload'))
        prefix = sys.prefix
        s1 = os.path.join(prefix, 'lib', 'site-packages')
        self.assertFalse(in_search_path(s1))
        python = 'python-%s.%s' % sys.version_info[:2]
        s2 = os.path.join(prefix, 'lib', python, 'site-packages')
        self.assertFalse(in_search_path(s2))
        import setuptools
        egg = os.path.dirname(setuptools.__file__)
        if '.egg' in egg:
            self.assertFalse(in_search_path(egg))

    def test_search_modules(self):
        from plone.reload.code import search_modules
        found = False
        we = 'plone' + os.sep + 'reload' + os.sep + '__init__.py'
        for f, m in search_modules():
            if we in f:
                found = True
                break
        self.assertTrue(found)

    def test_search_modules_eggs(self):
        from plone.reload.code import search_modules
        from plone.reload import config
        try:
            esp = config.EXCLUDE_SITE_PACKAGES
            config.EXCLUDE_SITE_PACKAGES = False
            found = False
            for f, m in search_modules():
                if '.egg' in f:
                    found = True
            self.assertTrue(found)
        finally:
            config.EXCLUDE_SITE_PACKAGES = esp


class TestTimes(unittest.TestCase):

    def test_get_mod_time(self):
        from plone.reload.code import get_mod_time
        tests = os.path.join(TESTS, '__init__.py')
        our_time = get_mod_time(tests)
        self.assertEqual(our_time, os.stat(tests)[8])

    def test_get_mod_time_compiled(self):
        from plone.reload.code import get_mod_time
        tests = os.path.join(TESTS, '__init__.py')
        tests_c = os.path.join(TESTS, '__init__.pyc')
        our_time = get_mod_time(tests_c)
        self.assertEqual(our_time, os.stat(tests)[8])

    def test_get_mod_times(self):
        from plone.reload.code import get_mod_times
        our_package = os.path.abspath(
            os.path.join(TESTS, os.pardir, '__init__.pyc'))
        times = get_mod_times()
        self.assertTrue(our_package in times)
        self.assertTrue(isinstance(times[our_package][1], types.ModuleType))

    def test_check_mod_times(self):
        from plone.reload.code import check_mod_times
        self.assertEqual(len(check_mod_times()), 0)

    def test_check_mod_times_change(self):
        from plone.reload.code import check_mod_times
        from plone.reload.code import MOD_TIMES
        our_package = os.path.abspath(
            os.path.join(TESTS, os.pardir, '__init__.pyc'))
        our_entry = MOD_TIMES[our_package]
        try:
            MOD_TIMES[our_package] = (our_entry[0] - 10, our_entry[1])
            self.assertEqual(len(check_mod_times()), 1)
        finally:
            MOD_TIMES[our_package] = our_entry

    def test_reload_code(self):
        from plone.reload.code import reload_code
        self.assertEqual(len(reload_code()), 0)

    def test_reload_code_change(self):
        from plone.reload.code import reload_code
        from plone.reload.code import MOD_TIMES
        our_package = os.path.abspath(
            os.path.join(TESTS, os.pardir, '__init__.pyc'))
        our_entry = MOD_TIMES[our_package]
        try:
            MOD_TIMES[our_package] = (our_entry[0] - 10, our_entry[1])
            self.assertEqual(len(reload_code()), 1)
        finally:
            MOD_TIMES[our_package] = our_entry



from App.config import getConfiguration
from zope.interface import implements
from zope.publisher.browser import BrowserView

from plone.reload.code import reload_code
from plone.reload.interfaces import IReload
from plone.reload.zcml import reload_zcml

HAS_CMF = True
try:
    from plone.reload.template import reload_template
except ImportError:
    HAS_CMF = False


class Reload(BrowserView):
    """Reload view.
    """
    implements(IReload)

    def __init__(self, context, request):
        BrowserView.__init__(self, context, request)
        self.message = None

    def __call__(self):
        action = self.request.form.get('action')
        if action is not None:
            if self.available():
                if action == 'code':
                    self.message = self.code_reload()
                elif action == 'zcml':
                    self.message = self.zcml_reload()
            if action == 'template':
                self.message = self.template_reload()
        return self.index()

    def available(self):
        if getConfiguration().debug_mode:
            return True
        return False

    def status(self):
        return self.message

    def template_reload_available(self):
        return HAS_CMF and not getConfiguration().debug_mode

    def template_reload(self):
        if HAS_CMF:
            reloaded = reload_template(self.context)
            if reloaded > 0:
                return '%s templates reloaded.' % reloaded
            return 'No templates reloaded.'
        return 'CMF is not installed. Templates cannot be reloaded.'

    def code_reload(self):
        if not self.available():
            return

        reloaded = reload_code()

        result = ''
        if reloaded:
            result += 'Code reloaded:\n\n'
            result += '\n'.join(reloaded)
        else:
            result = 'No code reloaded!'
        return result

    def zcml_reload(self):
        if not self.available():
            return

        # We always do an implicit code reload so we can register all newly
        # added classes.
        reloaded = reload_code()
        reload_zcml()

        # TODO Minimize all caches, we only really want to invalidate the
        # local site manager from all caches
        self.context._p_jar.db().cacheMinimize()
        result = ''
        if reloaded:
            result += 'Code reloaded:\n\n'
            result += '\n'.join(reloaded)
        else:
            result = 'No code reloaded!'
        result += '\n\nGlobal ZCML reloaded.'
        return result



# xreload.py.

"""Alternative to reload().

This works by executing the module in a scratch namespace, and then
patching classes, methods and functions. This avoids the need to
patch instances. New objects are copied into the target namespace.

Taken and extended from xreload as posted by Guido van Rossum:

    http://mail.python.org/pipermail/edu-sig/2007-February/007787.html

"""

import marshal
import imp
import sys
import types
import inspect

import zope.component


CLASS_STATICS = frozenset(["__dict__", "__doc__", "__module__", "__weakref__"])


class ClosureChanged(Exception):
    pass


class Reloader(object):
    """Reload a module in place, updating classes, methods and functions.

    Args:
      mod: a module object

    Returns:
      The (updated) input object itself.
    """

    def __init__(self, module):
        self.mod = module

    def reload(self):
        # Get the module name, e.g. 'foo.bar.whatever'
        modname = self.mod.__name__

        # Get the module namespace (dict) early; this is part of the type check
        modns = self.mod.__dict__
        # Parse it into package name and module name, e.g. 'foo.bar' and
        # 'whatever'
        i = modname.rfind(".")
        if i >= 0:
            pkgname, modname = modname[:i], modname[i + 1:]
        else:
            pkgname = None
        # Compute the search path
        if pkgname:
            # We're not reloading the package, only the module in it
            pkg = sys.modules[pkgname]
            path = pkg.__path__  # Search inside the package
        else:
            # Search the top-level module path
            pkg = None
            path = None  # Make find_module() uses the default search path
        # Find the module; may raise ImportError
        (stream, filename, (suffix, mode, kind)) = imp.find_module(
            modname, path)
        # Turn it into a code object
        try:
            # Is it Python source code or byte code read from a file?
            # XXX Could handle frozen modules, zip-import modules
            if kind not in (imp.PY_COMPILED, imp.PY_SOURCE):
                # Fall back to built-in reload()
                return reload(self.mod)
            if kind == imp.PY_SOURCE:
                source = stream.read()
                # PeterB: if we don't strip the source code and add newline we
                # get a SyntaxError even if `python $filename` is perfectly
                # happy.
                source = source.strip() + '\n'
                code = compile(source, filename, "exec")
            else:
                # I have no idea how to test this one
                code = marshal.load(stream)  # pragma NO COVER
        finally:
            if stream:
                stream.close()
        # Execute the code im a temporary namespace; if this fails, no changes
        tmpns = {'__name__': '%s.%s' % (pkgname, modname),
                 '__file__': filename,
                 '__doc__': modns['__doc__']}
        exec(code, tmpns)
        # Now we get to the hard part
        _update_scope(modns, tmpns)
        # Now update the rest in place
        for name in set(modns) & set(tmpns):
            modns[name] = self._update(modns[name], tmpns[name])
        # Done!
        return self.mod

    def _update(self, oldobj, newobj):
        """Update oldobj, if possible in place, with newobj.

        If oldobj is immutable, this simply returns newobj.

        Args:
          oldobj: the object to be updated
          newobj: the object used as the source for the update

        Returns:
          either oldobj, updated in place, or newobj.
        """
        if not isinstance(oldobj, type(newobj)):
            # Cop-out: if the type changed, give up
            return newobj

        new_module = getattr(newobj, '__module__', None)
        if new_module != self.mod.__name__:
            # Do not update objects in-place that have been imported.
            # Just update their references.
            return newobj

        if isinstance(newobj, zope.interface.interface.Specification):
            # XXX we can't update interfaces because their internal
            # data structures break. We'll have to implement the reload method
            # for those and patch it in.
            return oldobj
        if inspect.isclass(newobj):
            return _update_class(oldobj, newobj)
        elif inspect.isfunction(newobj):
            return _update_function(oldobj, newobj)

        # XXX Support class methods, static methods, other decorators
        # Not something we recognize, just give up
        # This line is currently not hit at all, since we only call this on
        # a module. It's pretty hard to have a non-function, non-class entity
        # in a module, which has a __module__ pointer to the module itself
        return newobj  # pragma NO COVER


def _closure_changed(oldcl, newcl):
    old = oldcl is None and -1 or len(oldcl)
    new = newcl is None and -1 or len(newcl)
    if old != new:
        return True
    if old > 0 and new > 0:
        for i in range(old):
            same = oldcl[i] == newcl[i]
            if not same:
                return True
    return False


def _update_scope(oldscope, newscope):
    oldnames = set(oldscope)
    newnames = set(newscope)
    # Add newly introduced names
    for name in newnames - oldnames:
        oldscope[name] = newscope[name]
    # Delete names that are no longer current
    for name in oldnames - newnames:
        if not name.startswith('__'):
            del oldscope[name]


def _update_function(oldfunc, newfunc):
    """Update a function object."""
    if _closure_changed(oldfunc.func_closure, newfunc.func_closure):
        raise ClosureChanged
    oldfunc.func_code = newfunc.func_code
    oldfunc.func_defaults = newfunc.func_defaults
    _update_scope(oldfunc.func_globals, newfunc.func_globals)
    # XXX What else?
    return oldfunc


def _update_method(oldmeth, newmeth):
    """Update a method object."""
    # XXX What if im_func is not a function?
    _update_function(oldmeth.im_func, newmeth.im_func)
    return oldmeth


def _update_class(oldclass, newclass):
    """Update a class object."""
    # XXX What about __slots__?
    olddict = oldclass.__dict__
    newdict = newclass.__dict__
    oldnames = set(olddict)
    newnames = set(newdict)
    for name in newnames - oldnames:
        setattr(oldclass, name, newdict[name])

    # Note: We do not delete attributes, because various ZCML directives,
    # grokkers and other wiring add class attributes during startup that
    # would get lost if we did this. Note that attributes will still be
    # overwritten if they've changed.
    #
    # for name in oldnames - newnames:
    #     delattr(oldclass, name)

    for name in oldnames & newnames - CLASS_STATICS:
        try:
            new = getattr(newclass, name)
            old = getattr(oldclass, name, None)
            if isinstance(new, types.MethodType):
                if isinstance(old, property) and not isinstance(new, property):
                    # Removing a decorator
                    setattr(oldclass, name, new.im_func)
                else:
                    _update_method(old, new)
            elif isinstance(new, types.FunctionType):
                # __init__ is a function
                _update_function(old, new)
            else:
                new2 = newdict.get(name)
                if new is not new2:
                    # Do we have some sort of descriptor? Set the underlying
                    # descriptor and not the result of the descriptor call
                    setattr(oldclass, name, new2)
                else:
                    # Fallback to just replace the item
                    setattr(oldclass, name, new)
        except ClosureChanged:
            # If the closure changed, we need to replace the entire function
            setattr(oldclass, name, new.im_func)

    return oldclass



from setuptools import setup

version = '2.0.1'

setup(
    name='plone.reload',
    version=version,
    description="Configuration and code reload without server restarts.",
    long_description=(open('README.rst').read() + '\n' +
                      open('CHANGES.rst').read()),
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 4.3",
        "Framework :: Plone :: 5.0",
        "Framework :: Zope2",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
      ],
    keywords='zope plone reload',
    author='Hanno Schlichting',
    author_email='hanno@hannosch.eu',
    url='https://pypi.python.org/pypi/plone.reload',
    license='BSD',
    packages=['plone', 'plone.reload'],
    namespace_packages=['plone'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools',
        'zope.component',
        'zope.interface',
        'zope.processlifetime',
        'zope.publisher',
        'zope.site',
        'zope.testing',
        'Zope2 >= 2.12',
    ],
    extras_require=dict(
      cmf=['Products.CMFCore'],
    ),
    entry_points="""
    [z3c.autoinclude.plugin]
    target = plone
    """,
    )



from zope.interface import Interface


class IReload(Interface):
    """Interface for the ZCML reload view.
    """

    def status():
        """Return a status text."""

    def code_reload():
        """Reload all changed code."""

    def zcml_reload():
        """Reprocess all global ZCML."""



from zope.component import getGlobalSiteManager
from zope.site.hooks import setSite
from zope.testing import cleanup

# BBB: Zope 2.12
try:
    from Zope2.App import zcml
    zcml  # pyflakes
except ImportError:
    from Products.Five import zcml


CORE_CLEANUPS = frozenset([
    'OFS.metaconfigure',
    'Products.Five.zcml',
    'Products.Five.eventconfigure',
    'Products.Five.fiveconfigure',
    'Products.Five.sizeconfigure',
    'zope.component.globalregistry',
    'zope.schema.vocabulary',
    'zope.security.management',
    'zope.security.checker',
    'zope.site.hooks',
    'Zope2.App.zcml',
])


def cleanups():
    registered = [c[0] for c in cleanup._cleanups]
    functions = []
    for r in registered:
        if r.__module__ not in CORE_CLEANUPS:
            functions.append(r)
    return functions


def reload_zcml():
    gsm = getGlobalSiteManager()
    old_gsm_dict = gsm.__dict__.copy()
    try:
        setSite(None)
        gsm.__init__(gsm.__name__)
        # Clean up
        for clean in cleanups():
            clean()
        # Reload all ZCML
        zcml._initialized = False
        zcml._context._seen_files.clear()
        zcml.load_site()
    except Exception as e:
        gsm.__init__(gsm.__name__)
        gsm.__dict__.clear()
        gsm.__dict__.update(old_gsm_dict)
        raise e
